import sys
import base64
import logging
import copy

import rsb
from rsb.converter import PredicateConverterList, Converter
from rsb import Event


class Forwarder(Converter):
    def __init__(self):
        super(Forwarder, self).__init__(bytearray, tuple, '.*')

    def serialize(self, data):
        return bytearray(data[1]), data[0]

    def deserialize(self, data, wireSchema):
        return wireSchema, data


class Bridge(object):
    basic_types = {'integer': int, 'float': float, 'string': lambda s: s, 'bool': bool}
    RSB_TO_WAMP = 1
    WAMP_TO_RSB = 2
    BIDIRECTIONAL = 3

    def __init__(self, rsb_scope, rsb_config, wamp, message_type, mode=BIDIRECTIONAL, wamp_scope=None):
        logging.info("register scopes:")
        self.mode = mode
        self.rsb_scope = rsb_scope
        self.wamp_scope = rsb_scope[1:].replace('/', '.') if wamp_scope is None else wamp_scope
        self.converter = None
        self.skipNext = False
        self.rsb_publisher = None
        self.rsb_listener = None
        logging.info("RSB Scope %s" % self.rsb_scope)
        logging.info("WAMP Scope is %s" % self.wamp_scope)
        self.wamp = wamp

        if message_type in Bridge.basic_types:
            self.wamp_callback = self.send_primitive_data
            self.rsb_callback = self.on_primitive_message
            self.rsb_type = Bridge.basic_types[message_type]
        else:
            self.wamp_callback = self.send_rst
            self.rsb_callback = self.on_bytearray_message
            self.rsb_type = str('.' + message_type)

        # RSB_TO_WAMP
        if mode % 2 > 0:
            logging.info('listening on rsb scope %s' % self.rsb_scope)
            self.rsb_listener = rsb.createListener(self.rsb_scope, config=rsb_config)
            self.rsb_listener.addHandler(self.rsb_callback)

        # WAMP_TO_RSB
        if mode > 1:
            logging.info('listening on wamp scope %s' % self.wamp_scope)
            self.wamp_listener = self.wamp.subscribe(self.on_wamp_message, self.wamp_scope)
            self.rsb_publisher = rsb.createInformer(self.rsb_scope, config=rsb_config)

    def on_bytearray_message(self, event):
        if 'wamp' in event.metaData.userInfos:
            logging.debug("received OWN rsb bytearray on %s, skipping..." % self.rsb_scope)
            return
        logging.debug('received rsb bytearray on %s' % self.rsb_scope)
        logging.debug('event length %d' % len(event.data[1]))
        logging.debug('sent to %s' % self.wamp_scope)
        try:
            msg = '\0' + base64.b64encode(event.data[1]).decode('ascii')
            self.wamp.publish(self.wamp_scope, msg)
        except Exception as e:
            logging.error(e)
            sys.exit(1)

    def on_primitive_message(self, event):
        if 'wamp' in event.metaData.userInfos:
            logging.debug("received OWN rsb primitive on %s, skipping..." % self.rsb_scope)
            return
        logging.debug("received rsb primitive [%s] on %s" % (str(event.data), self.rsb_scope))
        logging.debug("sent to %s" % self.wamp_scope)
        self.wamp.publish(self.wamp_scope, self.rsb_type(event.data))

    def send_rst(self, data):
        try:
            logging.info("send rst message to %s" % self.rsb_scope)
            binary_data = bytearray(base64.b64decode(data[1:]))
            event = Event(scope=self.rsb_scope,
                          data=(self.rsb_type, binary_data), type=tuple,
                          userInfos={'wamp':''})
            self.rsb_publisher.publishEvent(event)
        except Exception as e:
            logging.error(e)
            sys.exit(1)

    def send_primitive_data(self, data):
        try:
            logging.info("send primitive message [%s] message to %s" % (unicode(data), self.rsb_scope))
            self.rsb_publisher.publishData(self.rsb_type(data))
        except Exception as e:
            logging.error("Error while sending primitive data: %s" % str(e))
            sys.exit(1)

    def on_wamp_message(self, event):
        logging.debug('Received wamp message on %s' % self.wamp_scope)
        self.wamp_callback(event)

    def shutdown(self):
        logging.info("Shutting down bridge...")
        if self.rsb_listener:
            self.rsb_listener.deactivate()
        if self.rsb_publisher:
            self.rsb_publisher.deactivate()


def create_rsb_config():
    rsb_conf = copy.deepcopy(rsb.getDefaultParticipantConfig())
    trans = rsb_conf.getTransports()
    for t in trans:
        convs = rsb.convertersFromTransportConfig(t)

    conv = Forwarder()
    conv_list = PredicateConverterList(bytearray)
    conv_list.addConverter(conv,
                           dataTypePredicate=lambda data_type: issubclass(data_type, tuple),
                           wireSchemaPredicate=lambda wire_schema: wire_schema.startswith('.'))

    for t in trans:
        convs = rsb.convertersFromTransportConfig(t)
        for c in convs.getConverters().values():
            conv_list.addConverter(c,
                                   dataTypePredicate=lambda data_type, d_type=c.getDataType(): issubclass(data_type, d_type))
        c = rsb.converter.StringConverter()
        conv_list.addConverter(c,
                               dataTypePredicate=lambda data_type, d_type=c.getDataType(): issubclass(data_type, d_type))
        t.converters = conv_list
    return rsb_conf


class SessionHandler(object):

    def __init__(self, wamp_session, log_level=logging.WARNING):
        logging.basicConfig()
        logging.getLogger().setLevel(log_level)
        logging.getLogger("rsb").setLevel(logging.ERROR)

        self.wamp_session = wamp_session
        self.scopes = {}
        self.rsb_conf = create_rsb_config()

    def register_scope(self, rsb_scope, message_type):
        logging.info("trying to register on scope %s with message type %s" %
                     (rsb_scope, message_type))

        if rsb_scope not in self.scopes:
            b = Bridge(rsb_scope, self.rsb_conf, self.wamp_session, message_type)
            self.scopes[rsb_scope] = b
            return "Scope registered"
        return "Scope already exists"

    def quit(self):
        logging.info("quitting session...")
        for bridge in self.scopes.values():
            bridge.shutdown()



import logging
import os
from threading import Thread
import time

try:
    import asyncio
except ImportError:
    # Trollius >= 0.3 was renamed
    import trollius as asyncio

from autobahn.asyncio.wamp import ApplicationSession
from services import SessionHandler


class Ping(Thread):
    def __init__(self, wamp):
        Thread.__init__(self)
        self.running = True
        self.wamp = wamp

    def run(self):
        try:
            while self.running:
                logging.debug("ping")
                self.wamp.publish("com.wamp.ping", "ping")
                time.sleep(1)
        except Exception as e:
            logging.debug(e)
            raise e


class Component(ApplicationSession):

    @staticmethod
    def on_ping(event):
        logging.debug(event)

    @asyncio.coroutine
    def onJoin(self, details):
        # init members
        if os.environ.get('DEBUG') in ['1','True','true','TRUE']:
            log_level = logging.DEBUG
        else:
            log_level = logging.INFO
        logging.basicConfig()
        logging.getLogger().setLevel(log_level)
        self.session = SessionHandler(self, log_level)

        # register RPC
        reg = yield self.register(self.session.register_scope, 'service.displayserver.register')

        # setup ping
        sub = yield self.subscribe(self.on_ping, "com.wamp.ping")

        self.ping = Ping(self)
        self.ping.start()

        print 'kogniserver(asyncio) started...'

    def onLeave(self, details):
        self.ping.running = False
        while self.ping.isAlive():
            time.sleep(0.1)
        self.session.quit()
        print "kogniserver session left..."


def main_entry(use_ssl=False):
    from autobahn.asyncio.wamp import ApplicationRunner
    proto = "wss" if use_ssl else "ws"
    runner = ApplicationRunner(url=u"{0}://127.0.0.1:8181/ws".format(proto), realm=u"realm1")
    try:
        runner.run(Component)
    except KeyboardInterrupt or Exception:
        raise KeyboardInterrupt
    print "shutting down kogniserver..."


if __name__ == '__main__':
    main_entry()



import logging
import os
import argparse
import sys
from functools import partial

from .services import Bridge, create_rsb_config

try:
    import asyncio
except ImportError:
    # Trollius >= 0.3 was renamed
    import trollius as asyncio

from autobahn.asyncio.wamp import ApplicationSession


class Client(ApplicationSession):

    def __init__(self, config, scopes):
        self._scopes = [s.strip() for s in scopes.split(' ; ')]
        self.bridges = []
        self.config = create_rsb_config()
        super(Client, self).__init__(config)

    @asyncio.coroutine
    def onJoin(self, details):
        # init members
        if os.environ.get('DEBUG') in ['1', 'True', 'true', 'TRUE']:
            log_level = logging.DEBUG
        else:
            log_level = logging.INFO
        logging.basicConfig()
        logging.getLogger().setLevel(log_level)
        for mapping in self._scopes:
            self.bridges.append(Bridge(**self.parse_mapping(mapping)))
        print 'client(asyncio) connected...'

    def parse_mapping(self, mapping):
        source, direction, destination = mapping.split(' ')
        destination = None if '*' in destination else destination
        d = 0
        if direction.startswith('<'):
            d += Bridge.WAMP_TO_RSB
        if direction.endswith('>'):
            d += Bridge.RSB_TO_WAMP

        message_type = direction.translate(None, '<->')
        return {'rsb_scope': source, 'rsb_config': self.config, 'wamp': self, 'message_type': message_type,
                'mode': d, 'destination': destination}

    def onLeave(self, details):
        for bridge in self.bridges:
            bridge.shutdown()
        print "client session left..."


def main_entry(args=None):
    from autobahn.asyncio.wamp import ApplicationRunner

    parser = argparse.ArgumentParser()
    parser.add_argument('url', metavar='URL', type=unicode, help="URL to websocket of WAMP server")
    parser.add_argument('scopes', metavar='SCOPE', type=str,
                        help="scope mappings: /rsb/scope [<][-]message_type[-][>] wamp.scope")
    args = sys.argv[1:] if args is None else args
    args = parser.parse_args(args)
    runner = ApplicationRunner(url=args.url, realm=u"realm1")
    try:
        runner.run(partial(Client, scopes=args.scopes))
    except KeyboardInterrupt or Exception:
        raise KeyboardInterrupt
    print "shutting down client..."


if __name__ == '__main__':
    main_entry()



import argparse
from os import makedirs
from os.path import abspath, exists, join, dirname
import re
import subprocess
import time
import json
import threading
import socket
import sys

from .server import main_entry as async_main


def run_crossbar(config_path, keep_alive):
    ret = subprocess.call(['crossbar', 'status'])

    if ret == 0 and not keep_alive:
        subprocess.call(['crossbar', 'stop'])

    if ret != 0 or not keep_alive:
        subprocess.call(['crossbar', 'start', '--config=%s' % config_path])


def main_entry(args=None):

    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--force', help='overwrite config file if it already exists', action='store_true')
    parser.add_argument('-k', '--keep-alive', help='use existing crossbar instance', action='store_true')
    parser.add_argument('-c', '--config', help='location of the config file')
    parser.add_argument('-g', '--generate', help='only generate config file (with default options)', action='store_true')
    args = sys.argv[1:] if args is None else args
    args = parser.parse_args(args)

    pwd = abspath(__file__)
    elems = re.compile('[\\\\/]+').split(pwd)
    if 'site-packages' in elems:
        idx = elems.index('site-packages')
        elems = elems[:idx-2]
    else:
        elems = elems[:-1]
    prefix = join("/", *elems)

    config_path = join(prefix, 'etc/crossbar/config.json') if not args.config else args.config

    choice = 'n'
    if exists(config_path) is False:
        input_valid = False if not args.generate else True
        while not input_valid:
            choice = raw_input("config.json for crossbar does not exists. Should a default one be created? [y]/n:") or 'y'
            if choice not in ['y', 'n']:
                print("please enter 'y' or 'n'.")
            else:
                input_valid = True

    if choice in 'y' or args.force:
        protopath = join(prefix, 'share/rst0.12/proto')
        input_valid = False if not args.generate else True
        while not input_valid:
            protopath = raw_input("Location of proto-files? [%s]:" % protopath) or protopath
            if not exists(protopath) and not args.generate:
                print("%s does not exist!" % protopath)
            else:
                input_valid = True

        if exists(config_path) and not args.force:
            print "Config file already exists! Use --force to overwrite."
            return
        else:
            if not exists(dirname(config_path)):
                makedirs(dirname(config_path))

        with open(config_path, 'w') as target:
            j = json.loads(CONFIG_JSON)
            paths = j['workers'][0]['transports'][0]['paths']
            paths['/']['directory'] = join(prefix, paths['/']['directory'])
            if protopath:
                paths['proto']['directory'] = protopath
            else:
                del paths['proto']
            json.dump(j, target, indent=4)

    # In a dry generation run we can exit here
    if args.generate:
        return

    t1 = threading.Thread(target=run_crossbar, args=(config_path,args.keep_alive,))
    t1.setDaemon(True)
    t1.start()
    while not check_server('localhost', 8181):
        time.sleep(0.5)
    with open(config_path) as crossbar_config:
        j = json.load(crossbar_config)
        use_ssl = 'tls' in j['workers'][0]['transports'][0]['endpoint']
    async_main(use_ssl)

def check_server(address, port):
    # Create a TCP socket
    s = socket.socket()
    try:
        s.connect((address, port))
        #print "Connected to %s on port %s" % (address, port)
        s.close()
        return True
    except socket.error, e:
        #print "Connection to %s on port %s failed: %s" % (address, port, e)
        return False

if __name__ == '__main__':
    main_entry()



CONFIG_JSON = """
{
  "version": 2,
  "controller": {},
  "workers": [
    {
      "transports": [
        {
          "paths": {
            "ws": {
              "type": "websocket"
            },
            "/": {
              "directory": "var/www/kogniserver",
              "type": "static"
            },
            "proto": {
              "directory": "share/rst0.12/proto",
              "type": "static"
            }
          },
          "endpoint": {
            "type": "tcp",
            "port": 8181
          },
          "type": "web"
        }
      ],
      "type": "router",
      "options": {
        "pythonpath": [""]
      },
      "realms": [
        {
          "name": "realm1",
          "roles": [
            {
              "name": "anonymous",
              "permissions": [
                {
                  "uri": "",
                  "match": "prefix",
                  "allow": {
                    "call": true,
                    "register": true,
                    "publish": true,
                    "subscribe": true
                  },
                  "disclose": {
                    "caller": false,
                    "publisher": false
                  },
                  "cache": true
                }
              ]
            }
          ]
        }
      ]
    }
  ]
}
"""



from setuptools import setup, find_packages
from setuptools.command.install import install
import os
from os.path import join, exists, abspath
import json

setup(
    name='kogniserver',
    version='0.1.5',
    maintainer='Alexander Neumann',
    maintainer_email='aleneum@gmail.com',
    url='http://github.com/aleneum/kogniserver',
    description="Interface server of the KogniHome project",
    platforms=['Any'],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    tests_require=['nose>=1.3', 'coverage'],
    install_requires=['txaio', 'pyasn1', 'autobahn', 'crossbar', 'trollius', 'rsb-python'],
    entry_points={
        "console_scripts": [
            "kogniserver = kogniserver.adm:main_entry",
            "kogniclient = kogniserver.client:maint_entry"
        ]
    },
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Environment :: Console',
        'Environment :: No Input/Output (Daemon)',
        'Topic :: Communications',
        'Topic :: Home Automation',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Networking'
    ],
)
